/* mixer program for use with the joystick driver of linux-2.2.x
 *
 * Lars Tuber	(taeuber@informatik.hu-berlin.de)
 *
 */

#include <errno.h>
#include <linux/joystick.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/soundcard.h>
#include <string.h>
#include <unistd.h>

#ifndef JS_VERSION
#error wrong driver version (1.0 or greater expected)
#endif

#define LENGTH 5000	/* max length of .mixerc	*/
#define PERMS 0644	/* permissions of .mixerc	*/

#define S	1	/* used when -s is command		*/
#define C	2	/* used when -c is command		*/



void cal(int joyfd, char filename[]);	/* calibrates the resistors of joyfd	*/



int main (int argc, char *argv[])
{
    int joyfd,			/* filedescriptor for joystick device	*/
	rcfd,			/* filedescriptor for .mixerc		*/
	mixfd,			/* filedescriptor for mixer-device	*/
	i, j, k,		/* different usage			*/
	nread,			/* count read bytes			*/
	count,			/* stores some int-values of .mixerc	*/
	devmask,		/* aviable mixer devices		*/
	recmask,		/* aviable mixer devices for recording	*/
	recsrc,			/* record source settings		*/
	r[4] = {0,0,0,0},		/* points resistors to mixer device	*/
	b[4] = {0,0,0,0}, 	/* points buttons to mixer device	*/
	min[5] = {0,0,0,0,0},	/* min values of resistors (axes)	*/
	diff[5] = {0,0,0,0,0},	/* max-min values of resistors (axes)	*/
	line = 1;		/* counts lines in .mixerc file		*/

    unsigned char filename[100],		/* stores filename		*/
	 number_of_res,		/* number of resistors (axes)		*/
	 number_of_buts,	/* number of buttons			*/
	 buf[LENGTH],		/* stores content of .mixerc		*/
	 name[30],		/* stores name of mixer devices		*/
	 c = 0,			/* for single byte (r|b)		*/
	 mixername[50]="/dev/mixer",		/* stores mixer device name	*/
	 joyname[50]  ="/dev/js0",		/* stores js device name	*/
	 *names[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;	/* stores all possible mixer device names */

    struct js_event e;		/* event handler for joystick device	*/

    sprintf( filename, "%s/.mixerc", getenv("HOME"));

/*	ask for command line options	*/

    while ( -1 != (i = getopt (argc, argv, "hscvm:j:")))
    {
    	switch (i)
    	{
	    case 'v':	puts ("(c) 1999 Lars Tuber\t(taeuber@informatik.hu-berlin.de)\n"
			      "Version 0.2j\n");
			exit(0);

	    case 's':	c |= S;
			break;

	    case 'm':	sprintf (mixername, "%s", optarg);
			printf ("%s\n",mixername);
			break;

	    case 'j':	sprintf (joyname, "%s", optarg);
			printf ("%s\n",joyname);
			break;

	    case 'c':	c |= C;
			break;

	    default :	puts ("(c) 1999 Lars Tuber\t(taeuber@informatik.hu-berlin.de)\n"
			      "Version 0.2j\n"
			      "usage:\n"
			      "jmixer -h -m <mixer device> -s -j <joystick device> -c\n\n"
			      "\t-h for this help\n"
			      "\t-m <mixer device>\n"
			      "\t-s to have a look for aviable mixer devices\n"
			      "\t-j <joystick device>\n"
			      "\t-c for calibrating\n");
			exit(0);
	} /* switch */
    } /* while */

/*	open mixer device and read the needed values	*/

    if ( 0 > (mixfd = open ( mixername, O_RDWR)))
    {	sprintf (buf, "can't open device %s", mixername);
	perror (buf); 
	exit (-1);
    } /* if */
    
    if (ioctl(mixfd, SOUND_MIXER_READ_DEVMASK, &devmask) == -1)
    {	perror ("SOUND_MIXER_READ_DEVMASK");
	exit(-1);
    } /* if */

    if (ioctl(mixfd, SOUND_MIXER_READ_RECMASK, &recmask) == -1)
    {	perror ("SOUND_MIXER_READ_RECMASK");
	exit(-1);
    } /* if */

    if (ioctl(mixfd, SOUND_MIXER_READ_RECSRC, &recsrc) == -1)
    {	perror ("SOUND_MIXER_READ_RECSRC");
	exit(-1);
    } /* if */

/*	show the aviable mixer devices if '-s' is given	*/

    if (c & S)
    {
	puts ("\n(c) 1999 Lars Tuber\t(taeuber@informatik.hu-berlin.de)\n"
	      "aviable mixer devices:\n");
	for (j = 0; j < SOUND_MIXER_NRDEVICES; j++)
	    if ((1 << j) & devmask)
		printf ("%s ", names[j]);
	puts ("");
	exit(0);
    } /* if */

/*	open joystick device and read the number of resistors and buttons	*/

    if ( -1 == (joyfd = open (joyname, O_RDONLY)))
    {	sprintf (buf, "can't open device %s", joyname);
	perror (buf); 
	exit (-1); 
    } /* if */

    if ( -1 == ioctl (joyfd, JSIOCGAXES, &number_of_res))
    {	perror ("error: ioctl JSIOCGAXES");
	exit (-1);
    } /* if */
    else
	printf ("Number of resistors = %d\n", number_of_res);

    if ( -1 == ioctl (joyfd, JSIOCGBUTTONS, &number_of_buts))
    {	perror ("error: ioctl JSIOCBUTTONS");
	exit (-1);
    } /* if */
    else
	printf ("Number of buttons = %d\n", number_of_buts);

/*	calibrate the resistors if '-c' is given	*/

    if (c & C)
	cal(joyfd, filename);

/*	exclude the undefined resistors (axes)	*/

    for (i = 0; i < 4; i++)	r[i] = SOUND_MIXER_NRDEVICES + 1;

/*	try to open the ~/.mixerc file	*/

    if ( -1 == (rcfd = open ( filename, O_RDONLY)))
    {
	if (errno == ENOENT)
	    puts ("Please execute with '-c' for calibration and\n"
		  "define the channel control in '~/.mixerc'!");
	else
	    perror("can't open .mixerc");
    
        exit (-1); 
    } /* if */

/* read the .mixerc and connect the resistor and buttons to the mixer devices	*/

    if ( (nread = read (rcfd, buf, LENGTH)) == LENGTH)
    {	puts("~/.mixerc to long\n");
	exit(-1);
    } /* if */
    else
    {
	close (rcfd);

    	for ( i=0; i < nread; i++)
	{   j=0;
	    switch (buf[i])
	    {
		case '#' :  while ((buf[i] != '\n') && (i < nread))
				i++;
		case '\n':  line++;
			    break;

		case 'b' :
		case 'B' :
		case 'r' :
		case 'R' :  if (sscanf (&buf[i], "%1c %1d = %29[A-Za-z0-9]%n", &c, &count, name, &j) != 3)
			    {	printf("error in .mixerc line %d\n", line);
				exit(-1);
			    } /* if */
#ifdef DEBUG
			    printf("%c%d = %s\n", c, count, name);
#endif
			    i += j-1;
			    break;

		case 'c' :  if ((sscanf (&buf[i], "cal %1d = %d : %d%n", &count, &min[4], &diff[4], &j) != 3)
				|| (count < 0) 
				|| (count > 3))
			    {	printf("error in .mixerc line %d\n", line);
				exit(-1);
			    } /* if */
			    
			    diff[count] = diff[4] - (min[count] = min[4]);
#ifdef DEBUG
			    printf("cal%d = %d:%d\n", count, min[4], diff[4]);
#endif
			    i += j;
			    j = 0;
		default  :  break; 
	    } /* switch */
	    
	    if (j != 0)
	    {   for ( k = 0; k < SOUND_MIXER_NRDEVICES && strcmp(names[k], name); k++);
		if ((count < 0) || (count > 3) || (k >= SOUND_MIXER_NRDEVICES))
		{   printf("error in .mixerc line %d\n", line);
    		    exit(-1);
		} /* if */
	        else
	    	    switch (c)
		    {   case 'r':
		        case 'R':	r[count] = k;
					break;
			default :	b[count] = (1 << k) & recmask;
		    } /* switch */
	    } /* if */
	} /* for */
    } /* else */

/*	do the actual job	*/

    while (1)
    {
	if ( -1 == read(joyfd, &e, sizeof(struct js_event)))
        {   sprintf (buf, "can't read device %s", joyname);	
	    perror (buf);
	    exit (-1);
        } /* if */

	switch ( e.type & ~JS_EVENT_INIT)
	{	case JS_EVENT_BUTTON:
#ifdef DEBUG
					printf ("Button%1d   = %6d\r",e.number,e.value);
#endif
					if (e.value == 1)
					    recsrc |= b[e.number];
					else
					    recsrc &= ~b[e.number];
					    
					if (ioctl (mixfd, SOUND_MIXER_WRITE_RECSRC, &recsrc) == -1)
					{	perror("SOUND_MIXER_WRITE_RECSRC");
						exit (-1);
					} /* if */
					break;
					
		case JS_EVENT_AXIS  :	if ((1 << r[e.number]) & devmask)
					{
#ifdef DEBUG
						printf ("Resistor%1d = %6d\r", e.number, e.value);
#endif
						i = e.value - min[e.number];
						i |= ((i /= (diff[e.number]/100)) < 101 ? i : 100) << 8;
						if (ioctl (mixfd, MIXER_WRITE(r[e.number]), &i) == -1)
						{	perror("MIXER_WRITE");
							exit (-1);
						} /* if */
					} /* if */

	} /* switch */
	
    } /* while */

} /* main */


/*	calibrate the resistorsconnected to the joy-analog device	*/

void cal(int joyfd, char filename[])	/* calibrates the resistors (axes)	*/
{
    int  rcfd,				/* filedescriptor			*/
	 nread,				/* count read bytes			*/
	 flags = O_RDONLY,		/* how toopen the .mixerc		*/
	 k, j, dummy, count,
	 min[4] = {0,0,0,0},		/* min values of resistors (axes)	*/
	 max[4] = {0,0,0,0};		/* max values of resistors (axes)	*/

    unsigned char i, l,
		  res = ~15,		/* store the calibrated resistors	*/
		  buf[LENGTH],
		  newrc[LENGTH];	/* content of the recreated .mixerc	*/

    struct js_event e;			/* event handler for joystick device	*/

    puts("Please move the resistors to the minimum and to the maximum!"
         "Afterwards press a joystick button!\n");

    while (1)
    {
        if ( -1 == read(joyfd, &e, sizeof(struct js_event)))
	{
	    perror ("can't read joystick device");
	    exit (-1);
	} /* if */

	switch (e.type)
        {
    	    case JS_EVENT_AXIS  :	
#ifdef DEBUG
			        printf ("Resistor%d = %d\n",e.number,e.value);
#endif
				res |= 1 << e.number;
				if (min[e.number] > e.value)
				    min[e.number] = e.value;
				else
				    if (max[e.number] < e.value)
					max[e.number] = e.value;
				break;

	    case JS_EVENT_BUTTON:
	    
			/*	open the .mixerc file if it exists otherwise create a new one	*/

			        while ((rcfd = open (filename, flags, PERMS)) == -1)
			        {
				    if (flags & O_CREAT)
				    {
					perror("can't create .mixerc");
				        exit (-1); 
				    } /* if */
	
				    flags = O_CREAT | O_RDWR | O_TRUNC | O_SYNC;
				} /* while */

			        if ( (nread = read (rcfd, buf, LENGTH)) >= LENGTH - 100)
			        {
				    puts("~/.mixerc to long\n");
				    exit(-1);
			        } /* if */

				/*	is .mixerc is created new - then it is readonly	*/

				if (!(flags &= O_CREAT))
				    close(rcfd);
#ifdef DEBUG
				printf("nread=%d res=%d\n", nread, res);
#endif
				/* delete the old calibrated values	*/

				for (count = j = 0; j < nread; j++)
				{
				    if ((sscanf (&buf[j], "cal %1d = %d : %d%n", &l, &dummy, &dummy, &k) == 3) && (res & (1 << l)))
					j += k;
				    else
				        if (!((newrc[count - 1] == '\n') && (buf[j] == '\n')))
					    newrc[count++] = buf[j];
				} /* for */
				newrc[count]='\0';
#ifdef DEBUG
				printf("count=%d newrc=%s\n",count,newrc);
#endif
				/* if .mixerc is not created then do it now	and truncate it	*/

				if (!flags)
				    if ( -1 == creat (filename, PERMS))
				    {    
				        perror("can't create .mixerc");
					exit (-1); 
				    } /* if */

				/* write the undeletet text	*/
					
				if (count)
				    write (rcfd, newrc, count);

				/* write the new calibrating values	*/

				for (i = 0; i < 4; i++)
				{	if ((1 << i) & res )
				    	{   
					    sprintf(buf, "\ncal%d=%d:%d", i, min[i], max[i]);
					    printf("cal %d = %s\n",i,buf);
					    if ( -1 == write (rcfd, buf, strlen(buf)))
					    {   perror ("error writing .mixerc");
						exit(-1);
					    } /* if */

					} /* if */

				} /* for */
				
				close (rcfd);
				exit(0);
				    
	} /* switch */

    } /* while */
    	
} /* cal */